package ru.amse.vorobiev.lce.interpreter;

import ru.amse.vorobiev.lce.circuit.ICircuit;
import ru.amse.vorobiev.lce.elements.*;
import ru.amse.vorobiev.lce.elements.impl.GateType;
import ru.amse.vorobiev.lce.elements.impl.InGate;
import ru.amse.vorobiev.lce.elements.impl.OutGate;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Queue;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

public class Interpreter implements IInterpreter {
	private ICircuit myCircuit;
	private Set<InGate> myInGates;
	private final List<Queue<IGate>> myQueues;
	private Map<IPlug, Boolean> myVisitedPlugs;
		
	public Interpreter(ICircuit circuit) {
		myVisitedPlugs = new HashMap<IPlug, Boolean>();
		myCircuit = circuit;
		myInGates = myCircuit.getInGates();
		myQueues = new ArrayList<Queue<IGate>>();
		for (IGate inGate : myInGates) {
			Queue<IGate> queue = new LinkedList<IGate>();
			queue.add(inGate);;
			myQueues.add(queue);
		}
	}
	
	public Interpreter() {
		myVisitedPlugs = new HashMap<IPlug, Boolean>();
		myCircuit = null;
		myInGates = null;
		myQueues = new ArrayList<Queue<IGate>>();
	}
	
	public void setUp(ICircuit circuit) {
		myCircuit = circuit;
		myInGates = myCircuit.getInGates();
		myQueues.clear();
		for (IGate inGate : myInGates) {
			Queue<IGate> queue = new LinkedList<IGate>();
			queue.add(inGate);;
			myQueues.add(queue);
		}
	}
	
	public void reset() {
		int i = 0;
		for (IGate inGate : myInGates) {
			Queue<IGate> queue = myQueues.get(i);
			queue.clear();
			queue.add(inGate);
			i++;
		}
		List<IPlug> plugsToRemove = new LinkedList<IPlug>();
		for (IPlug plug : myVisitedPlugs.keySet()) {
			if (plug.getGate().getType() != GateType.IN) {
				plugsToRemove.add(plug);
			}
		}
		for (IPlug plug : plugsToRemove) {
			myVisitedPlugs.remove(plug);
		}
	}
	
	public void setInputs(InGate gate, boolean[] inputs) {
		if (gate.getType() == GateType.IN) {
			for (int i = 0; i < gate.getOutputCount(); i++) {
				myVisitedPlugs.put(gate.getOutput(i), inputs[i]);
			}
		} else {
			throw new IllegalArgumentException();
		}
	}
	
	private void calculateOutputs(IGate gate) {
		GateType type = gate.getType();
		if (type == GateType.IN || type == GateType.OUT) {
			throw new IllegalArgumentException();
		}
		boolean value = myVisitedPlugs.get(gate.getInput(0));
		switch (type) {
			case AND: {
				for (int i = 0; i < gate.getInputCount(); i++) {
					boolean input = myVisitedPlugs.get(gate.getInput(i));
					value &= input;
				}
				break;
			}
			case OR: {
				for (int i = 0; i < gate.getInputCount(); i++) {
					boolean input = myVisitedPlugs.get(gate.getInput(i));
					value |= input;
				}
				break;
			} 
			case XOR: {
				for (int i = 0; i < gate.getInputCount(); i++) {
					boolean input = myVisitedPlugs.get(gate.getInput(i));
					value ^= input;
				}
				break;
			} 
			case NOT: {
				value = !value; 
				break;
			}
		}
		for (int i = 0; i < gate.getOutputCount(); i++) {
			myVisitedPlugs.put(gate.getOutput(i), value);
		}
	}
	
	private boolean isReady(IGate gate) {
		for (int i = 0; i < gate.getInputCount(); i++) {
			if (myVisitedPlugs.get(gate.getInput(i)) == null) {
				return false;
			}
		}
		return true;
	}
	
	public void step() {
		for (int w = 0; w < myQueues.size(); w++) {
			Queue<IGate> queue = myQueues.get(w);
			int size = queue.size();
			for (int i = 0; i < size; i++) {
				IGate gate = queue.poll();
				if (gate.getType() != GateType.OUT) {
					if (gate.getType() != GateType.IN) {
						calculateOutputs(gate);
					}
					for (int j = 0; j < gate.getOutputCount(); j++) {
						IPlug output = gate.getOutput(j);
						IPlug input = output.getWire().getTo();
						myVisitedPlugs.put(input, myVisitedPlugs.get(output));
						IGate nextGate = input.getGate();
						if (isReady(nextGate)) {
							queue.add(nextGate);
						}
					}
				}
			}
		}
	}
	
	public void interprete() {
		for (Queue<IGate> queue : myQueues) {
			while (!queue.isEmpty()) {
				step();
			}
		}
	}
	
	public Set<InGate> getInGates() {
		return myCircuit.getInGates();
	}
	
	public Set<OutGate> getOutGates() {
		return myCircuit.getOutGates();
	}

	public Map<IPlug, Boolean> getVisitedPlugs() {
		return myVisitedPlugs;
	}

	public boolean allInGatesAreReady() {

		boolean isReadyGate = true;
		for (InGate gate : myInGates) {
			for (int i = 0; i < gate.getOutputCount(); i++) {
				IPlug plug = gate.getOutput(i);
				if (myVisitedPlugs.get(plug) == null) {					
					isReadyGate = false;
				}
			}
			if (isReadyGate) {
				return true;
			}
			isReadyGate = true;
		}
		return false;
	}
	
/*	public Queue<IGate> getQueue() {
*		return myQueue;
*	}
*/
}
