package ru.amse.vorobiev.lce.ui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.HashSet;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.xml.stream.XMLStreamException;

import ru.amse.vorobiev.lce.circuit.Circuit;
import ru.amse.vorobiev.lce.circuit.CircuitChangeListener;
import ru.amse.vorobiev.lce.circuit.ICircuit;
import ru.amse.vorobiev.lce.interpreter.IInterpreter;
import ru.amse.vorobiev.lce.io.*;
import ru.amse.vorobiev.lce.ui.tools.ITool;
import ru.amse.vorobiev.lce.ui.tools.impl.*;
import ru.amse.vorobiev.lce.view.ICircuitView;
import ru.amse.vorobiev.lce.view.IGateView;
import ru.amse.vorobiev.lce.view.impl.CircuitView;

public class JEditorFrame extends JFrame {

	private static final long serialVersionUID = 1L;
	
	private JEditor myEditor;
	private final StepAction myStepAction;
	private final InterpreteAction myInterpreteAction;
	private final ResetInterpreterAction myResetInterpreterAction;
	private final NewFileAction myNewFileAction;
	private final OpenFileAction myOpenAction;
	private final SaveAction mySaveAction;
	private final SaveAsAction mySaveAsAction;
	private final DeleteSelectedGateAction myDeleteAction;
	private final JScrollPane myErrorPane;
	private final JTable myErrorTable;
	private AbstractAction myCurrentTool;
	private File myCurrentFile;
	private boolean hasChanged;

	public JEditorFrame() {
		hasChanged = false;
		myCurrentFile = null;
		setSize(800, 600);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		addWindowListener(new FrameClosingListener(this));
		setTitle("New circuit - LCE");
		myEditor = new JEditor();
		myEditor.addChangeListener(new EditorChangeListener(this));
		ICircuit circuit = myEditor.getView().getCircuit();
		circuit.addChangeListener(new CircuitChangeListener(circuit, this));

		myOpenAction = new OpenFileAction(myEditor);
		mySaveAction = new SaveAction(myEditor);
		mySaveAction.setEnabled(false);
		mySaveAsAction = new SaveAsAction(myEditor);
		myDeleteAction = new DeleteSelectedGateAction(myEditor);
		myNewFileAction = new NewFileAction(myEditor);
				
		JMenuBar menuBar = new JMenuBar();
		JMenu fileMenu = new JMenu("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		
		fileMenu.add(myNewFileAction);
		fileMenu.add(myOpenAction);
		fileMenu.add(mySaveAction);
		fileMenu.add(mySaveAsAction);
		fileMenu.add(new JSeparator());
		fileMenu.add(new ExitAction());
		
		JMenu editMenu = new JMenu("Edit");
		
		editMenu.add(myDeleteAction);
		
		menuBar.add(fileMenu);
		menuBar.add(editMenu);
		this.setJMenuBar(menuBar);
		
		JToolBar toolbar = new JToolBar();
		toolbar.setFloatable(false);
		toolbar.add(myNewFileAction);
		toolbar.add(myOpenAction);
		toolbar.add(mySaveAction);
		myStepAction = new StepAction(null);
		myInterpreteAction = new InterpreteAction(null);
		myResetInterpreterAction = new ResetInterpreterAction(null);
		setInterpreterActive(false);
		toolbar.add(myStepAction);
		toolbar.add(myInterpreteAction);
		toolbar.add(myResetInterpreterAction);
		AbstractAction selectionTool = new SetSelectionToolsAction(myEditor);
		JToggleButton selection = new JToggleButton(selectionTool);
		JToggleButton and = new JToggleButton(new AndCreationToolsAction(myEditor));
		JToggleButton or = new JToggleButton((new OrCreationToolsAction(myEditor)));
		JToggleButton not = new JToggleButton(new NotCreationToolsAction(myEditor));
		JToggleButton xor = new JToggleButton(new XorCreationToolsAction(myEditor));
		JToggleButton in = new JToggleButton(new InCreationToolsAction(myEditor));
		JToggleButton out = new JToggleButton(new OutCreationToolsAction(myEditor));
		
		ButtonGroup buttonGroup = new ButtonGroup();
		buttonGroup.add(selection);
		buttonGroup.add(and);
		buttonGroup.add(not);
		buttonGroup.add(or);
		buttonGroup.add(xor);
		buttonGroup.add(in);
		buttonGroup.add(out);
		
		toolbar.add(selection);
		toolbar.add(and);
		toolbar.add(or);
		toolbar.add(not);
		toolbar.add(xor);
		toolbar.add(in);
		toolbar.add(out);
		getContentPane().add(toolbar, BorderLayout.NORTH);
		selectionTool.actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, ""));
		
		ErrorTableModel tableModel = new ErrorTableModel();
		myErrorTable = new JTable(tableModel);
		ListSelectionModel selectionModel = myErrorTable.getSelectionModel();
		selectionModel.addListSelectionListener(new ErrorSelectionListener(tableModel, myEditor));
		myErrorTable.setSelectionModel(selectionModel);
		myErrorTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		myErrorTable.setRowSelectionAllowed(true);
		myErrorTable.setColumnSelectionAllowed(false);
		
		
		JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		JScrollPane editorScrollPane = new JScrollPane(myEditor);
		myErrorPane = new JScrollPane(myErrorTable);
		splitPane.setTopComponent(editorScrollPane);
		splitPane.setBottomComponent(myErrorPane);
		splitPane.setResizeWeight(1.0);
		splitPane.setDividerLocation(400);
		
		getContentPane().add(splitPane);
	}
	
	public JEditor getEditor() {
		return myEditor;
	}
	
	public void setChanged(boolean changed) {
		String title = this.getTitle();
		if (hasChanged != changed) {
			if (!changed) {
				title = title.replaceFirst("~", "");
				mySaveAction.setEnabled(false);
			} else {
				title = "~" + title;
				mySaveAction.setEnabled(true);
			}
			this.setTitle(title);
			hasChanged = changed;
		}
	}
	
	public boolean hasChanged() {
		return hasChanged;
	}
	
	public void save() {
		mySaveAction.actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, ""));
	}
	
	public void setInterpreterActive(boolean isActive) {
		myInterpreteAction.setEnabled(isActive);
		myStepAction.setEnabled(isActive);
		myResetInterpreterAction.setEnabled(isActive);
	}
	
	public void setSaverActive(boolean isActive) {
		mySaveAsAction.setEnabled(isActive);
	}
	
	public JTable getErrorTable() {
		return myErrorTable;
	}
	
	public JScrollPane getErrorPane() {
		return myErrorPane;
	}
	
	public void setInterpreter(IInterpreter interpreter) {
		myInterpreteAction.setInterpreter(interpreter);
		myStepAction.setInterpreter(interpreter);
		myResetInterpreterAction.setInterpreter(interpreter);
	}
	
	private void setCurrentTool(AbstractAction tool) {
		myEditor.clearSelection();
		if (myCurrentTool != null) {
			if (myCurrentTool instanceof DeleteSelectedGateAction) {
				myEditor.clearSelection();
			}
			myCurrentTool.setEnabled(true);
		}
		myCurrentTool = tool;
//		myCurrentTool.setEnabled(false);
        this.repaint();
	}
	
	public int promptForSaving() {
		int answer = JOptionPane.showConfirmDialog(this, "Do you want to save changes?", "Close application", JOptionPane.YES_NO_CANCEL_OPTION);
		if (answer == JOptionPane.YES_OPTION) {
			this.save();
		}
		return answer;
	}
			
	private class InterpreteAction extends AbstractAction {

		private static final long serialVersionUID = 1L;
		private IInterpreter myInterpreter;

		public InterpreteAction(IInterpreter interpreter) {
			myInterpreter = interpreter;
			this.putValue(NAME, "interprete");
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Interprete.gif")));
		}
		
		public void setInterpreter(IInterpreter interpreter) {
			myInterpreter = interpreter;
		}
		
		public void actionPerformed(ActionEvent e) {
			myInterpreter.interprete();
			JEditorFrame.this.repaint();
		}
	}
	
	private class StepAction extends AbstractAction {

		private static final long serialVersionUID = 1L;
		private IInterpreter myInterpreter;
		
		public StepAction(IInterpreter interpreter) {
			myInterpreter = interpreter;
			this.putValue(NAME, "step");
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Step.gif")));
		}
		
		public void setInterpreter(IInterpreter interpreter) {
			myInterpreter = interpreter;
		}

		public void actionPerformed(ActionEvent e) {
			myInterpreter.step();
			JEditorFrame.this.repaint();
		}
	}
	
	private class ResetInterpreterAction extends AbstractAction {

		private static final long serialVersionUID = 1L;
		private IInterpreter myInterpreter;
		
		public ResetInterpreterAction(IInterpreter interpreter) {
			myInterpreter = interpreter;
			this.putValue(NAME, "reset");
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Stop.gif")));
		}
		
		public void setInterpreter(IInterpreter interpreter) {
			myInterpreter = interpreter;
		}

		public void actionPerformed(ActionEvent e) {
			myInterpreter.reset();
			JEditorFrame.this.repaint();
		}
	}
	
	private class NewFileAction extends AbstractAction {

		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		
		public NewFileAction(JEditor editor) {
			myEditor = editor;
			this.putValue(NAME, "New");
			this.putValue(AbstractAction.ACCELERATOR_KEY, 
					KeyStroke.getKeyStroke(KeyEvent.VK_N, KeyEvent.CTRL_DOWN_MASK));
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/New.gif")));
		}

		public void actionPerformed(ActionEvent e) {
			int answer = -1;
			if (JEditorFrame.this.hasChanged()) {
				answer = JEditorFrame.this.promptForSaving();
			}
			if (answer != JOptionPane.CANCEL_OPTION) {
				JEditorFrame.this.setTitle("New circuit - LCE");
				ICircuit circuit = new Circuit();
				circuit.addChangeListener(new CircuitChangeListener(circuit, JEditorFrame.this));
				circuit.fireCircuitChanged();
				myEditor.setView(new CircuitView(circuit));
				myEditor.setSize(new Dimension(0,0));
				myEditor.setPreferredSize(new Dimension(0,0));
				JEditorFrame.this.setChanged(false);
				myEditor.repaint();
			}
		}
	}
	
	private class OpenFileAction extends AbstractAction {

		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		
		public OpenFileAction(JEditor editor) {
			myEditor = editor;
			this.putValue(NAME, "Open");
			this.putValue(AbstractAction.ACCELERATOR_KEY, 
					KeyStroke.getKeyStroke(KeyEvent.VK_O, KeyEvent.CTRL_DOWN_MASK));
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Open.gif")));
		}
		
		public void actionPerformed(ActionEvent e) {
			int answer = -1;
			if (JEditorFrame.this.hasChanged()) {
				answer = JEditorFrame.this.promptForSaving();
			}
			if (answer != JOptionPane.CANCEL_OPTION) {
				JFileChooser fileChooser = new JFileChooser();
				fileChooser.setFileFilter(new EditorFileFilter());
				fileChooser.showOpenDialog(JEditorFrame.this);
				File file = fileChooser.getSelectedFile();
				if (file != null) {
					CircuitLoader loader = new CircuitLoader();
					try {
						loader.loadCircuit(file);
						ICircuitView view = loader.getCircuitView();
						myEditor.setView(view);
						myEditor.ensureSize();
						ICircuit circuit = view.getCircuit();
						circuit.addChangeListener(new CircuitChangeListener(circuit, JEditorFrame.this));
						circuit.fireCircuitChanged();
						JEditorFrame.this.setChanged(false);
						JEditorFrame.this.repaint();
						myEditor.getView().getCircuit().getInterpreter().setUp(view.getCircuit());
					} catch (WrongFileStructureException ex) {
						JOptionPane.showMessageDialog(JEditorFrame.this, "File has invalid structure");
					} catch (WrongFileFormatException ex) {
						JOptionPane.showMessageDialog(JEditorFrame.this, "File has invalid xml format");
					}
					myCurrentFile = file;
					JEditorFrame.this.setTitle(file.getName() + " - LCE");
				}
			}
		}
	}
	
	private class SaveAsAction extends AbstractAction{

		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		
		public SaveAsAction(JEditor editor) {
			myEditor = editor;
			this.putValue(NAME, "Save as");
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Save.gif")));
		}

		public void actionPerformed(ActionEvent e) {
			JFileChooser fileChooser = new JFileChooser();
			fileChooser.setFileFilter(new EditorFileFilter());
			fileChooser.showSaveDialog(JEditorFrame.this);
			File file = fileChooser.getSelectedFile();
			if (file != null) {
				String path = file.getAbsolutePath();
				if (!path.endsWith(".xml")) {
					file = new File(path + ".xml");
				}
				CircuitSaver saver = new CircuitSaver();
				try {
					saver.saveCircuitToFile(myEditor.getView(), file);
					JEditorFrame.this.setChanged(false);
					JEditorFrame.this.myCurrentFile = file;
					JEditorFrame.this.setTitle(file.getName() + " - LCE");
				} catch (FileNotFoundException e1) {
					JOptionPane.showMessageDialog(JEditorFrame.this, "File was not found");
				} catch (XMLStreamException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}
		}
		
	}
	
	private class SaveAction extends AbstractAction{

		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		
		public SaveAction(JEditor editor) {
			myEditor = editor;
			this.putValue(NAME, "Save");
			this.putValue(AbstractAction.ACCELERATOR_KEY, 
					KeyStroke.getKeyStroke(KeyEvent.VK_S, KeyEvent.CTRL_DOWN_MASK));
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Save.gif")));
		}

		public void actionPerformed(ActionEvent e) {
			if (myCurrentFile != null) {
				CircuitSaver saver = new CircuitSaver();
				try {
					saver.saveCircuitToFile(myEditor.getView(), myCurrentFile);
					JEditorFrame.this.setChanged(false);
				} catch (FileNotFoundException e1) {
					JOptionPane.showMessageDialog(JEditorFrame.this, "File was not found");
				} catch (XMLStreamException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			} else {
				JEditorFrame.this.mySaveAsAction.actionPerformed(e);
			}
		}
		
	}
	
	private class ExitAction extends AbstractAction{

		private static final long serialVersionUID = 1L;
		
		public ExitAction() {
			this.putValue(NAME, "Exit");
			this.putValue(AbstractAction.ACCELERATOR_KEY, 
					KeyStroke.getKeyStroke(KeyEvent.VK_W, KeyEvent.CTRL_DOWN_MASK));
		}

		public void actionPerformed(ActionEvent e) {
			if (JEditorFrame.this.hasChanged()) {
				int answer = JEditorFrame.this.promptForSaving();
				if (answer != JOptionPane.CANCEL_OPTION) {
					System.exit(0);
				}
			} else {
				System.exit(0);
			}
		}
	}
	
	private class DeleteSelectedGateAction extends AbstractAction {
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		
		public DeleteSelectedGateAction(JEditor editor) {
			myEditor = editor;
			this.putValue(NAME, "Delete");
			this.putValue(AbstractAction.ACCELERATOR_KEY, KeyStroke.getKeyStroke("DELETE"));
		}

		public void actionPerformed(ActionEvent e) {
			Set<IGateView> gates = myEditor.getSelectedGates();
			for (IGateView gateView : gates) {
				myEditor.getView().removeViewByElement(gateView.getGate());
				myEditor.getView().getCircuit().removeGate(gateView.getGate());
			}
			myEditor.clearSelection();
			myEditor.repaint();
		}	
	}
	
	private class SetSelectionToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final Set<ITool> mySelectionTools;

		public SetSelectionToolsAction(JEditor editor) {
			myEditor = editor;
			mySelectionTools = new HashSet<ITool>();
			mySelectionTools.add(new WireSelectionTool(myEditor));
			mySelectionTools.add(new GateDraggingTool(myEditor));
			mySelectionTools.add(new SettingInputsTool(myEditor));
			mySelectionTools.add(new GateSelectionTool(myEditor));
			mySelectionTools.add(new AddingInputOutputTool(myEditor));
			mySelectionTools.add(new DeletingInputOutputTool(myEditor));
			this.putValue(AbstractAction.SMALL_ICON, new ImageIcon(this.getClass().getResource("icons/Select.gif")));
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTools(mySelectionTools);
			setCurrentTool(this);
		}
		
	}
	
	private class AndCreationToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final ITool myAndCreationTool;

		public AndCreationToolsAction(JEditor editor) {
			myEditor = editor;
			myAndCreationTool = new AndGateCreationTool(myEditor);
			this.putValue(NAME, "AND");
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTool(myAndCreationTool);
			setCurrentTool(this);
		}	
	}
	
	private class OrCreationToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final ITool myOrCreationTool;

		public OrCreationToolsAction(JEditor editor) {
			myEditor = editor;
			myOrCreationTool = new OrGateCreationTool(myEditor);
			this.putValue(NAME, "OR");
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTool(myOrCreationTool);
			setCurrentTool(this);
		}	
	}
	
	private class NotCreationToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final ITool myNotCreationTool;

		public NotCreationToolsAction(JEditor editor) {
			myEditor = editor;
			myNotCreationTool = new NotGateCreationTool(myEditor);
			this.putValue(NAME, "NOT");
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTool(myNotCreationTool);
			setCurrentTool(this);
		}	
	}
	
	private class XorCreationToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final ITool myXorCreationTool;

		public XorCreationToolsAction(JEditor editor) {
			myEditor = editor;
			myXorCreationTool = new XorGateCreationTool(myEditor);
			this.putValue(NAME, "XOR");
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTool(myXorCreationTool);
			setCurrentTool(this);
		}	
	}
	
	private class InCreationToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final ITool myInCreationTool;

		public InCreationToolsAction(JEditor editor) {
			myEditor = editor;
			myInCreationTool = new InGateCreationTool(myEditor);
			this.putValue(NAME, "IN");
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTool(myInCreationTool);
			setCurrentTool(this);
		}	
	}
	
	private class OutCreationToolsAction extends AbstractAction {
		
		private static final long serialVersionUID = 1L;
		private JEditor myEditor;
		private final ITool myOutCreationTool;

		public OutCreationToolsAction(JEditor editor) {
			myEditor = editor;
			myOutCreationTool = new OutGateCreationTool(myEditor);
			this.putValue(NAME, "OUT");
		}
		
		public void actionPerformed(ActionEvent e) {
			myEditor.setCurrentTool(myOutCreationTool);
			setCurrentTool(this);
		}	
	}
	
}
